-- PC Parallel Port Model for XSB board 
-- this file simulates APSI activety based on the apsi.tst file 
-- generated by apsi.exe in vhdlsim mode
-- spsi.tst (binary file)

-- author: Ernest Jamro

library ieee;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

entity epp_model is
	generic (T: time:= 41.6ns); -- CPLD clock perion = 24MHz
  port(pp_dck: out std_logic; -- data clock
	  pp_dwr: out STD_LOGIC; -- write data
    pp_drd: in std_logic); -- read data
end epp_model;


architecture epp_model_arch of epp_model is
  constant ScriptName : string := "apsi.tst";   -- nazwa pliku z danymi do symulacji
  constant OutFileName : string := "src/apsi.out"; -- all readblock data are written there
  -- types 
  type FileType is file of integer;
  type command_type is (wr_adr, wr_data, rd_data, sleep, wait_adr, wait_rd, unknown, finish);
  type WaitAdrType is array (3 downto 0) of integer;
-----------------------------------------------------------
-- this function converts std_logic_vector to integer
function conv_int(arg: std_logic_vector) return integer is --unsigned input data
    variable ret: integer;
  begin
	ret:= 0;
  	for i in arg'high downto arg'low loop
		ret:= ret*2;
		if arg(i)= '1' then ret:= ret+1; end if;
	end loop;
    return ret;
  end;
-------------------------------------------------------------------------------------------  
  -- skryp file
  file ScriptFile: FileType;
  
 -- read script file -- return integer which is written on 1 or 3 bytes (NrBytes)
procedure ReadScript(variable ScriptDataIn: inout integer; 
          variable ScriptNrReadBytes: inout integer;
          variable read_data: out integer) is 
  Variable ret: integer; -- return value
  Variable DataIn: integer;
  Variable NrReadBytes: integer;
begin
  NrReadBytes:= ScriptNrReadBytes;
  DataIn:= ScriptDataIn;
  if ScriptNrReadBytes = 0 then -- read from file next 4 bytes
	READ(ScriptFile, DataIn);
	NrReadBytes:= 4; -- 4 bytes read from the file
  end if;
  ret:= DataIn rem 256;
  -- shift DataIn 8 bits right (divide by 256)
  if DataIn < 0 then -- if DataIn is negative then division by 256 causes wrong rounding
    ScriptDataIn:= (DataIn - ret) / 256; -- proper rounding for negative numbers
  else
	ScriptDataIn:= DataIn  / 256; -- no rounding problems for possitive numbers
  end if;
  ScriptNrReadBytes:= NrReadBytes - 1; 
  read_data:= ret;
end ReadScript;

----------------------------------------------------------------------------------
-- waitbit instruction (script instruction) Read Parammeters
procedure RdWaitBit(variable ScriptDataIn: inout integer; 
  variable ScriptNrReadBytes: inout integer;
	variable WaitAdr: out WaitAdrType; -- address (4 cycles that must be continuosly read)
	variable WaitMask: out std_logic_vector(7 downto 0); -- mask
	variable WaitBitOne: out boolean) -- for for bit one (true) or zero (false)
	 is
	
	variable read_data: integer;
begin
     ReadScript(ScriptDataIn, ScriptNrReadBytes, read_data); -- bit waitbit0 or waitbit1 command
     if read_data = 0 then
	     WaitBitOne:= false;
     else WaitBitOne:= true;
	   end if;
	  for i in 0 to 3 loop
	     ReadScript(ScriptDataIn, ScriptNrReadBytes, read_data); -- read MSB of the address
         WaitAdr(i):= read_data;
	   end loop;
	  ReadScript(ScriptDataIn, ScriptNrReadBytes, read_data); -- mask
    WaitMask:= conv_std_logic_vector(read_data, 8);
end RdWaitBit;

-----------------------------------------------
--- Check if read data from parallel port satisfy wait condition (wait should be finnished)
procedure CheckWaitBit(	variable WaitMask: in std_logic_vector(7 downto 0); -- mask
    variable WaitBitOne: in boolean; -- wait for bit one (true) or zero (false)
	  variable pp_data: in std_logic_vector(7 downto 0); -- data read from  the parallel port
	  variable next_wait_stage: inout integer) is -- command may be changed if wait is finished

  variable BitOne: boolean:= false;
begin
  for i in 0 to 7 loop
	if pp_data(i)='1' and WaitMask(i)='1' then
		BitOne:= true;
	end if;
  end loop;
  if (BitOne= true and  WaitBitOne = true) or (BitOne=false and WaitBitOne=false) then
	 next_wait_stage:= -1; -- finish the wait state
  else -- still wait condition
	 next_wait_stage:= 0; -- start reading from the beginning
  end if;
end CheckWaitBit;
	  
-----------------------------------------------------------------------------
-- Write Data to epp.out file
  file DataOutFile: FileType;
 
procedure FileOut(constant read_data: std_logic_vector(7 downto 0);
    variable FileOutNrBytes: inout integer;
    variable FileOutData: inout std_logic_vector(0 to 31)) is
  Variable write_data: std_logic_vector(31 downto 0);
  variable write_data_int: integer;
begin
  write_data:= FileOutData;
  write_data(FileOutNrBytes*8+7 downto FileOutNrBytes*8):= read_data;
  if FileOutNrBytes = 3 then
	write_data_int:= conv_int(write_data);
	WRITE(DataOutFile, write_data_int);
	FileOutNrBytes:= 0;
  else
	FileOutNrBytes:= FileOutNrBytes + 1;
	FileOutData:= write_data;
  end if;
end FileOut;

---------------------------------------------------------------------------------

-- Find Sim data (sim_data, sim_wr, sim_adr and command) from script file
Procedure SimData (variable sim_data: out integer; -- simulation data
  variable command: out command_type; variable ScriptDataIn: inout integer;
  variable ScriptNrReadBytes: inout integer) is

    variable read_data: integer;
  	variable cmd: integer; -- command
  begin
	ReadScript(ScriptDataIn, ScriptNrReadBytes, cmd);
   case cmd is
	when 1 | 2 => -- transmit address (data) - 1 Byte
	  ReadScript(ScriptDataIn, ScriptNrReadBytes, read_data);
	  if cmd=1 then
      command:= wr_adr;
    else 
      command:= wr_data;
    end if;
	  sim_data:=  read_data;
	when 3=> -- readbyte and write it to out file
	  command:= rd_data;
	when 10 => -- sleep (read the slep time)
	  command:= sleep;
	  ReadScript(ScriptDataIn, ScriptNrReadBytes, read_data);
	  sim_data:= read_data;
	when 11 => --wait bit
	  command:= wait_adr;
	when 255=> -- finish
	  command:= finish;
	when others => command:= unknown;
  end case;
end SimData;

  signal command: command_type; -- command read from epp.tst
  signal pp_data: std_logic_vector(7 downto 0); -- simalation data

begin

process 
 begin
  FILE_OPEN (ScriptFile, ScriptName, READ_MODE);
  FILE_OPEN (DataOutFile, OutFileName, WRITE_MODE);
  wait for 100000ms; -- wait forever
end process;

process
   variable stage: integer:= 0; -- stage of EPP bus
   variable cmd: command_type; -- command
   variable sleep_time: integer;
   variable ScriptNrReadBytes: integer:= 0; -- number of Bytes Read by ReadScript function form the script file and not returned by the function 
   variable ScriptDataIn: integer:= 0; -- data (integer read from the script 
   variable FileOutNrBytes: integer:= 0;
   variable FileOutData: std_logic_vector(0 to 31);
   variable pp_data_int: integer; -- data to be transfered o PP (integer value)
   variable sim_data: std_logic_vector(7 downto 0); -- data to be transffered on Parallel Port
   variable wait_stage: integer:= -1; -- stage of the wait instruction: -1 - no wait stage; 0,1,2,3 - address write stage, 4- data stage
   variable WaitAdr: WaitAdrType; -- address (4 cycles that must be continuosly read)
   variable WaitMask: std_logic_vector(7 downto 0); -- mask
   variable WaitBitOne, WaitResult: boolean; -- for for bit one (true) or zero (false)
   variable bit_data: std_logic_vector(7 downto 0); -- read write data
  begin	    
     if stage=0 and wait_stage<0 then -- read command from the file
       SimData(pp_data_int, cmd, ScriptDataIn, ScriptNrReadBytes);
       pp_dck<= '0'; 
		   pp_dwr<= '0';
       wait for 110ns; -- initial wait state
     end if; 
     command<= cmd;
     wait for 3*T;   
     
     case cmd is	
		  -- waitbit, unknown, finish
      when wr_adr | wr_data | wait_adr => -- write address or data
        if cmd = wait_adr then -- wait instruction is executed
          if wait_stage=-1 then -- the first wait cycle
            RdWaitBit(ScriptDataIn, ScriptNrReadBytes, WaitAdr, WaitMask, WaitBitOne);
            wait_stage:= 0;
           end if;
           pp_data_int:= WaitAdr(wait_stage);
         end if; -- end wait instruction             
        
        if stage=0 then -- the first 4 bits
          bit_data:= conv_std_logic_vector(pp_data_int, 8);
          pp_data<= bit_data;
          stage:= 1;
        else -- the last MSB 4 bits
          bit_data(3 downto 0):= bit_data(7 downto 4);
          stage:= 0;
        end if;
        if cmd = wr_adr or cmd = wait_adr then
          pp_dwr<= '1'; -- address
        else pp_dwr<= '0'; -- data
        end if;
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= '0'; -- write
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= bit_data(0); wait for T;-- LSB bit of 4-bit address/data
        while pp_drd='0' loop -- wait until FPGA is ready
          wait for T;
        end loop;          
        pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= bit_data(1); 
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= bit_data(2);
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= bit_data(3);
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        
        if stage=0 and wait_stage>=0 then -- wait instruction (whole address byte written)
          wait_stage:= wait_stage + 1;
          if wait_stage=4 then -- the last address instruc
            cmd:= wait_rd;
          end if;
        end if;
      
      when rd_data | wait_rd => -- read data
        pp_dwr<= '0'; -- data
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0';
        pp_dwr<= '1'; -- read
        wait for T; pp_dck<='1'; wait for T; pp_dck<='0'; wait for T; 
        pp_dwr<= '0'; -- do not care
        while pp_drd='0' loop -- wait until FPGA is ready
          wait for T;
        end loop;          
        pp_dck<='1'; wait for T; pp_dck<='0'; wait for T;
        bit_data(5 downto 0):= bit_data(7 downto 2); -- shift 2 bits left
        bit_data(6):= pp_drd;
        pp_dck<='1'; wait for T; pp_dck<='0'; wait for T;
        bit_data(7):= pp_drd;
        if stage<3 then -- not full read yet
          stage:= stage + 1;
        else -- the last read
          stage:= 0;
          pp_data<= bit_data;
          if wait_stage<0 then -- normal read
            FileOut(bit_data, FileOutNrBytes, FileOutData);
          else -- waitbit instruction
            CheckWaitBit(	WaitMask, WaitBitOne, bit_data, wait_stage); -- chect if the wait condition was met
            cmd:= wait_adr;
          end if;  
        end if;
        pp_dck<='1'; wait for T; pp_dck<='0';

      when sleep => -- sleep command
        wait for pp_data_int * 1ms;

      when finish=> -- the end
      		wait for 20*T;
          FileOut("11111111", FileOutNrBytes, FileOutData); -- write to fill 4 bytes
					FileOut("11111111", FileOutNrBytes, FileOutData); -- write to fill 4 bytes
					FileOut("11111111", FileOutNrBytes, FileOutData); -- write to fill 4 bytes
					assert false
                report "O.K. End of simulation (by epp_model, no further commands to execute)."
  	            severity failure;
          wait for 10000 ms;
		  when others => null;
		end case;
  end process;    
		 	
 
end epp_model_arch;



